@@ -1,5 +1,5 @@
 /* This file is part of GNU cflow
-   Copyright (C) 1997-2021 Sergey Poznyakoff
+   Copyright (C) 1997-2024 Sergey Poznyakoff
  
    GNU cflow is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -594,6 +594,14 @@ expression()
 	  switch (tok.type) {
 	  case ';':
 	       return;
+	  case PARM_WRAPPER:
+	       if (skip_balanced('(', ')', 0) == -1) {
+		    file_error(_("unexpected end of file in expression"),
+			       NULL);
+		    return;
+	       }
+	       putback();
+	       break;
 	  case LBRACE:
 	  case LBRACE0:
 	  case RBRACE:
@@ -637,12 +645,23 @@ expression()
 	       break;
 	  case '(':
 	       /* maybe typecast */
-	       if (nexttoken() == TYPE || tok.type == STRUCT)
-		    skip_to(')');
-	       else {
-		    putback();
+	       if (nexttoken() == TYPE || tok.type == STRUCT) {
+		    if (skip_balanced('(', ')', 1) == -1) {
+			 file_error(_("unexpected end of file in expression"),
+				    NULL);
+			 return;
+		    }
+		    if (tok.type == LBRACE || tok.type == LBRACE0) {
+			 if (skip_balanced('{', '}', 1) == -1) {
+			      file_error(_("unexpected end of file in expression"),
+					 NULL);
+			      return;
+			 }
+		    }
+	       } else {
 		    parens_lev++;
 	       }
+	       putback();
 	       break;
 	  case ')':
 	       parens_lev--;
@@ -700,6 +719,10 @@ parse_function_declaration(Ident *ident, int parm)
 int
 fake_struct(Ident *ident)
 {
+     while (tok.type == QUALIFIER &&
+	    (strcmp(tok.token, "const") == 0 ||
+	     strcmp(tok.token, "volatile") == 0))
+	  nexttoken();
      ident->type_end = -1;
      if (tok.type == STRUCT) {
 	  if (nexttoken() == IDENTIFIER) {
@@ -760,8 +783,8 @@ parse_variable_declaration(Ident *ident, int parm)
      default:
 	  if (verbose) 
 	       file_error(_("expected `;'"), &tok);
-	  /* FIXME: should putback() here */
-	  /* FALLTHRU */
+	  /* FIXME: should putback() here? */
+	  break;
      case ';':
 	  break;
      case ',':
@@ -772,11 +795,24 @@ parse_variable_declaration(Ident *ident, int parm)
 	  goto again;
      case '=':
 	  nexttoken();
-	  if (tok.type == LBRACE || tok.type == LBRACE0)
+	  switch (tok.type) {
+	  case LBRACE:
+	  case LBRACE0:
 	       initializer_list();
-	  else
+	       goto select;
+
+	  case RBRACE:
+	  case RBRACE0:
+	       if (verbose)
+		    file_error(_("expected expression"), &tok);
+	       break;
+
+	  default:
 	       expression();
-	  goto select;
+	       goto select;
+	  }
+	  break;
+
      case LBRACE0:
      case LBRACE:
 	  func_body();
@@ -895,6 +931,7 @@ dcl(Ident *idptr)
 			       NULL);
 		    return 1;
 	       }
+	       putback();	       
 	  } else if (tok.type == IDENTIFIER) {
 	       int type;
 	       
@@ -1066,7 +1103,7 @@ func_body()
 	       parse_declaration(&ident, 0);
 	       break;
 	  case LBRACE0:
-	  case '{':
+	  case LBRACE:
 	       level++;
 	       break;
 	  case RBRACE0:
@@ -1080,7 +1117,7 @@ func_body()
 	       }
 	       /* else: */
 	       /* FALLTHRU */
-	  case '}':
+	  case RBRACE:
 	       delete_autos(level);
 	       level--;
 	       break;
@@ -1088,10 +1125,10 @@ func_body()
 	       if (verbose)
 		    file_error(_("unexpected end of file in function body"),
 			       NULL);
-	       caller = NULL;
-	       return;
+	       goto end;
 	  }
      }
+end:
      caller = NULL;
 }
 
