@@ -1,5 +1,5 @@
 /* This file is part of GNU cflow
-   Copyright (C) 1997-2021 Sergey Poznyakoff
+   Copyright (C) 1997-2024 Sergey Poznyakoff
 
    GNU cflow is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -20,18 +20,26 @@
 
 static Hash_table *symbol_table;
 
+static struct linked_list *start_symbol_list;
+static struct linked_list *target_symbol_list;
 static struct linked_list *static_symbol_list;
 static struct linked_list *auto_symbol_list;
 static struct linked_list *static_func_list;
 static struct linked_list *unit_local_list;
 
 static void
-append_symbol(struct linked_list **plist, Symbol *sp)
+symbol_unlink_from_list(Symbol *sp)
 {
      if (sp->entry) {
 	  linked_list_unlink(sp->entry->list, sp->entry);
 	  sp->entry = NULL;
      }
+}
+
+static void
+append_symbol(struct linked_list **plist, Symbol *sp)
+{
+     symbol_unlink_from_list(sp);
      if (!data_in_list(sp, *plist)) {
 	  linked_list_append(plist, sp);
 	  sp->entry = (*plist)->tail;
@@ -93,6 +101,7 @@ install(char *name, int flags)
      memset(sym, 0, sizeof(*sym));
      sym->type = SymUndefined;
      sym->name = name;
+     sym->ord = -1;
 
      tp = xmalloc(sizeof(*tp));
      tp->sym = sym;
@@ -143,19 +152,16 @@ ident_change_storage(Symbol *sp, enum storage storage)
 	  append_symbol(&auto_symbol_list, sp);
 	  break;
      default:
+	  symbol_unlink_from_list(sp);
 	  break;
      }
+     
      sp->storage = storage;
 }
 
-Symbol *
-install_ident(char *name, enum storage storage)
+void
+init_ident(Symbol *sp, enum storage storage)
 {
-     Symbol *sp;
-
-     sp = install(name, 
-                  storage != AutoStorage ? 
-                     INSTALL_CHECK_LOCAL : INSTALL_DEFAULT);
      sp->type = SymIdentifier;
      sp->arity = -1;
      sp->storage = ExternStorage;
@@ -166,6 +172,17 @@ install_ident(char *name, enum storage storage)
      sp->caller = sp->callee = NULL;
      sp->level = -1;
      ident_change_storage(sp, storage);
+}
+
+Symbol *
+install_ident(char *name, enum storage storage)
+{
+     Symbol *sp;
+
+     sp = install(name, 
+                  storage != AutoStorage ? 
+                     INSTALL_CHECK_LOCAL : INSTALL_DEFAULT);
+     init_ident(sp, storage);
      return sp;
 }
 
@@ -289,6 +306,7 @@ struct collect_data {
      Symbol **sym;
      int (*sel)(Symbol *p);
      size_t index;
+     size_t count;
 };
 
 static bool
@@ -299,9 +317,10 @@ collect_processor(void *data, void *proc_data)
      Symbol *s;
      for (s = t->sym; s; s = s->next) {
 	  if (cd->sel(s)) {
-	       if (cd->sym)
-		    cd->sym[cd->index] = s;
-	       cd->index++;
+	       if (cd->index == cd->count) {
+		    cd->sym = x2nrealloc(cd->sym, &cd->count, sizeof(cd->sym[0]));
+	       }
+	       cd->sym[cd->index++] = s;
 	  }
      }
      return true;
@@ -313,8 +332,10 @@ collect_list_entry(void *item, void *proc_data)
      Symbol *s = item;
      struct collect_data *cd = proc_data;
      if (cd->sel(s)) {
-	  cd->sym[cd->index] = s;
-	  cd->index++;
+	  if (cd->index == cd->count) {
+	       cd->sym = x2nrealloc(cd->sym, &cd->count, sizeof(cd->sym[0]));
+	  }
+	  cd->sym[cd->index++] = s;
      }
      return 0;
 }
@@ -324,12 +345,9 @@ collect_symbols(Symbol ***return_sym, int (*sel)(Symbol *p),
 		size_t reserved_slots)
 {
      struct collect_data cdata;
-     size_t size;
      
-     size = hash_get_n_entries(symbol_table)
-	     + linked_list_size(static_func_list)
-	     + linked_list_size(unit_local_list);
-     cdata.sym = xcalloc(size + reserved_slots, sizeof(*cdata.sym));
+     cdata.sym = NULL;
+     cdata.count = 0;
      cdata.index = 0;
      cdata.sel = sel;
      hash_do_for_each(symbol_table, collect_processor, &cdata);
@@ -408,4 +426,109 @@ move_parms(int level)
 	  }
      }
 }
+
+Symbol *
+install_starter(char *name)
+{
+     Symbol *sp = install(name, 0);
+     sp->flag = symbol_start;
+     if (!data_in_list(sp, start_symbol_list))
+	  linked_list_append(&start_symbol_list, sp);
+     return sp;
+}
+
+void
+set_default_starter(void)
+{
+     if (!linked_list_head(start_symbol_list))
+	  install_starter("main");
+}
+
+void
+clear_starters(void)
+{
+     linked_list_destroy(&start_symbol_list);
+}
+
+Symbol *
+first_starter(void *itr)
+{
+     struct linked_list_entry *p = linked_list_head(start_symbol_list);
+     while (p) {
+	  Symbol *sym = p->data;
+	  if (sym->type != SymUndefined) {
+	       *(void**)itr = p->next;
+	       return sym;
+	  }
+	  p = p->next;
+     }
+     *(void**)itr = NULL;
+     return NULL;
+}
+
+Symbol *
+next_starter(void *itr)
+{
+     struct linked_list_entry *p;
+
+     if (!itr)
+	  return NULL;
+     p = *(void**)itr;
+     while (p) {
+	  Symbol *sym = p->data;
+	  if (sym->type != SymUndefined) {
+	       *(void**)itr = p->next;
+	       return sym;
+	  }
+	  p = p->next;
+     }
+     *(void**)itr = NULL;
+     return NULL;
+}
+
+Symbol *
+install_target(char *name)
+{
+     Symbol *sp = install(name, 0);
+     sp->flag = symbol_target;
+     if (!data_in_list(sp, target_symbol_list))
+	  linked_list_append(&target_symbol_list, sp);
+     return sp;
+}
+
+static void
+mark_callers(Symbol *sym)
+{
+     struct linked_list_entry *p;
+
+     if (sym->active)
+	  return;
+     sym->active = 1;
+     for (p = linked_list_head(sym->caller); p; p = p->next) {
+	  mark_callers((Symbol*)p->data);
+     }
+     sym->visible = 1;
+     sym->active = 0;
+}
+
+/* Eliminate all paths not leading to selected targets */
+void
+eliminate_non_targets(void)
+{
+     struct linked_list_entry *p;
+     
+     p = linked_list_head(target_symbol_list);
+     
+     if (!p)
+	  return;
+     
+     for (; p; p = p->next) {
+	  Symbol *sym = p->data;
+	  if (sym->flag == symbol_target) {
+	       mark_callers(sym);
+	  }
+     }
+     output_visible = 1;
+}
+
 
